/* written by Sven Petai <hadara@bsd.ee> */
/* 05.2003 */
#include <limits.h> 
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>

#include "table.h"

char *_errstr = NULL;

struct field {
	char type;
	char *name;
	char len; /* length of field in bytes */
	struct field *next;
};
/* it will probably be populated by lots of information in the
 * future like mutexes etc.
 */
struct table_handler {
	char *tablename;
	struct field *fieldlist;
	unsigned long int last_oid; /* greatest currently used auto increment record id */
	unsigned int struct_size; /* this is a size of the structure for comunicating with user applications, defined by schema */
	FILE *stream;
	long start; /* offset of the first record from the beginning of the file */
	long read_pos; /* start of the last record read */
	long end; /* position of EOF */
};		

int create_table(FILE *schema) {
	/* creates empty table file and copies schema into it */
	char fnamebuf[_POSIX_PATH_MAX];
	FILE *tstream;
	int i;
	long int size;
	char *schema_buf;
	unsigned char namelen;
	struct stat table;

	/* read name of the table */
	fread(&namelen, 1, 1, schema);
	if (namelen+strlen(DBDIR) >= _POSIX_PATH_MAX) {
		set_error("Path too long");
		return(-1);
	}
	sprintf(fnamebuf, DBDIR);
	fread(fnamebuf+strlen(DBDIR), 1, namelen, schema);
	fnamebuf[namelen+strlen(DBDIR)] = 0;
#ifdef DEBUG
	printf("Trying %s\n", fnamebuf);
#endif
	if (-1 != stat(fnamebuf, &table)) {
		set_error("Table exists. tablenames have to be unique");
		return(-1);
	}	
	/* how to avoid this race condition ? */
	if (NULL == (tstream = fopen(fnamebuf, "wb+"))) {
		set_error("Failed to open file");
		return(-1);
	}
	/* get length of schema */
	fseek(schema, 0L, SEEK_END);
	size = ftell(schema);
	rewind(schema);
	if (!(schema_buf = (char*)malloc(size * sizeof(char)))) {
		set_error("Not enough memory");
		return(-1);
	}
#ifdef DEBUG
	printf("Schema length is %d\n", size);
#endif
	if (size != fread(schema_buf, 1, size, schema)) {
		set_error("Failed to read schema in");
		return(-1);
	}
	/* write out schema to new table file */
	//fprintf(tstream, "%u",(unsigned int)size);
	fwrite(&size, 1, sizeof(unsigned int), tstream);
	fwrite(schema_buf, 1, size, tstream);	
	fclose(tstream);	
}

struct table_handler *open_datafile(char *table_name) {
	/* opens table file and reads schema into memory
	 * returns NULL when error occurs otherwise unsigned int is returned
	 * that identifies database for further actions like del_rec etc.
	 */
	char fnamebuf[_POSIX_PATH_MAX];
	struct table_handler *table_entry;
	unsigned int schema_length;
	char tmp;
	struct field *p;
	
#ifdef DEBUG
	printf("in open_datafile()\n");
#endif
	if (!(table_entry = (struct table_handler*)malloc(sizeof(struct table_handler)))) {
		set_error("No memory for table datastruct");	
		return(NULL);
	}
	table_entry->tablename = strdup(table_name);
	sprintf(fnamebuf, DBDIR);
	sprintf(fnamebuf+strlen(DBDIR), "%s", table_entry->tablename);
	if (!(table_entry->stream = fopen(fnamebuf, "a+"))) {
#ifdef DEBUG
		printf("DB doesn't exist: %s\n", fnamebuf);
#endif
		set_error("DB file doesn't exist");
		return(NULL);
	}
	rewind(table_entry->stream);
	fread(&schema_length, sizeof(unsigned int), 1, table_entry->stream);
	/* we know tablename allready, move position indicator over it */
	fseek(table_entry->stream, (long)(strlen(table_entry->tablename)+sizeof(char)), SEEK_CUR);
	/* read in last used oid */
	fread(&table_entry->last_oid, sizeof(unsigned long int), 1, table_entry->stream);
#ifdef DEBUG
	printf("schema length is: %u\n", schema_length);
	printf("last used oid is: %lu\n", table_entry->last_oid);
#endif
	table_entry->struct_size = 0;
	/* at least one field must always exists */
	p = table_entry->fieldlist = (struct field*)malloc(sizeof(struct field));
	for (;;) {
		/* this gives length of the field name */
		fread(&tmp, sizeof(char), 1, table_entry->stream);
		if (tmp) {
			p->name = (char*)malloc(sizeof(char)*tmp+1);
			/* read fieldname */
			fread(p->name, sizeof(char), tmp, table_entry->stream);
			p->name[tmp] = 0; /* terminate string */
			/* fieldtype */
			fread(&p->type, sizeof(char), 1, table_entry->stream);		
			/* field length */
			fread(&p->len, sizeof(char), 1, table_entry->stream);
#ifdef DEBUG
			printf("length of namefield: %d\n", tmp);
			printf("fieldname: %s\n", p->name);
			printf("fieldtype: %c\n", p->type);
			printf("fieldlength: %d\n", p->len);
#endif
			if (p->type == 'v') 
				table_entry->struct_size += sizeof(char*);
			else if (p->type == 'd')
				table_entry->struct_size += sizeof(double);
			else if (p->type == 'f')
				table_entry->struct_size += sizeof(float);
			else if (p->type == 't')
				table_entry->struct_size += sizeof(long);
			else if (p->type == 'i')
				table_entry->struct_size += sizeof(int);
		} else {
			set_error("Schema corrupted: 0 length fieldname found");
			return(NULL);
		}
		if (ftell(table_entry->stream) < schema_length) {
			/* vaata kas on veel kirjeid */
			p->next = (struct field*)malloc(sizeof(struct field));	
			p = p->next;
			p->next = NULL;
		} else {
			break;
		}
	}
	p->next = NULL;
	table_entry->start = ftell(table_entry->stream);
	fseek(table_entry->stream, 0L, SEEK_END);
	table_entry->end = ftell(table_entry->stream);
	fseek(table_entry->stream, table_entry->start, SEEK_SET);
	table_entry->read_pos = 0;
#ifdef DEBUG
	printf("END: %ld START: %ld\n", table_entry->end, table_entry->start);
#endif
	return(table_entry);
}

#ifdef DEBUG
void show_schema(struct table_handler *handler) {
	struct field *p;

	printf("In show_schema()\n");
	printf("tablename: %s\nlast_used_oid: %ld\n", handler->tablename, handler->last_oid);
	p = handler->fieldlist;
	while (p) {
		printf("\tfield: %s type:%c length:%d\n", p->name, p->type, p->len);
		p = p->next;
	}
}
#endif

int close_datafile(struct table_handler *handler) {
	struct field *p, *old;

	p = handler->fieldlist;
	while (p) {	
		old = p;
		p = p->next;
		free(old->name);
		free(old);
	}
	free(handler->tablename);
	free(handler);
#ifdef DEBUG
	printf("DATAFILE CLOSED\n");
#endif
	return(1);
}

long unsigned int add_rec(struct table_handler *tablehandle, void *data_orig) {
	/* adds new record 
	 * returns 0 when error occurs, otherwise oid of the new record 
	 */
	void *record_header, *start;
	char *data;
	struct field *p;
	unsigned long int oid;
	unsigned int length = 0;
	unsigned int local_length = 0;
	unsigned char tmp;
	char *c_ptr;
	void *buf;
	register int i, j;

	data = (char*)data_orig;
	if (-1 == fseek(tablehandle->stream, 0L, SEEK_END)) {
		perror("fseek failed");	
		exit(EXIT_FAILURE);
	}
	/* first while{} calculates length of the datastruct - 
	 * we don't know it beforehand because we have varchars
	 * I have to calculate it here because length has to be writen out before data
	 */
	p = tablehandle->fieldlist;
	while (p) {
		j = 0;
		switch (p->type) {
			case 'v':
				memcpy(&c_ptr, data, sizeof(char*));
#ifdef DEBUG
				printf("%s\n", c_ptr);
#endif
				length += strlen(c_ptr) + 1;
				data += sizeof(char*);
				break;
			case 'i':
				length += sizeof(int);
				data += sizeof(int);
				break;
			case 'd':
				length += sizeof(double);
				data += sizeof(double);
				break;
			case 'f':
				length += sizeof(float);
				data += sizeof(float);
				break;
			case 't':
				length += sizeof(long int);
				data += sizeof(long int);
				break;
			default:
				set_error("unknown type");
				return(0);
		}
		p = p->next;
	}			
#ifdef DEBUG
	printf("In add_rec()\n");
	printf("Length of the new record is %u\n", length);
#endif
	if(!fwrite(&length, sizeof(int), 1, tablehandle->stream)) {
		if (ferror(tablehandle->stream)) {
			printf("error on\n");
		} else {
			perror("jama");
		}	
		exit(3);
	}
	tmp = 0;
	fwrite(&tmp, sizeof(char), 1, tablehandle->stream);
	tablehandle->last_oid++; 
	fwrite(&tablehandle->last_oid, sizeof(long int), 1, tablehandle->stream);
	p = tablehandle->fieldlist;
	data = data_orig;
	while (p) {
		switch (p->type) {
			case 'v':
				memcpy(&c_ptr, data, sizeof(char*));
#ifdef DEBUG
				printf("I:%s\n", c_ptr);
#endif
				tmp = strlen(c_ptr);
				if (!fwrite(&tmp, sizeof(char), 1, tablehandle->stream)) {
					perror("dsfds");	
					exit(2);
				}
				fwrite(c_ptr, sizeof(char), tmp, tablehandle->stream);
				data += sizeof(char*);
				break;
			case 'i':
				fwrite(data, sizeof(int), 1, tablehandle->stream);
				data += sizeof(int);
				break;
			case 'd':
				fwrite(data, sizeof(double), 1, tablehandle->stream);
				data += sizeof(double);
				break;
			case 'f':
				fwrite(data, sizeof(float), 1, tablehandle->stream);
				data += sizeof(float);
				break;
			case 't':
				fwrite(data, sizeof(long int), 1, tablehandle->stream);
				data += sizeof(long int);
				break;
			default:
				set_error("unknown type");
				return(0);
		}
		p = p->next;
	}
	fflush(tablehandle->stream);				
	tablehandle->end += length;
}

int read_rec(unsigned long int recno, void **recbuf, struct table_handler *tablehandle) {
	int length;
	long unsigned int cur_oid;

	/* return 0 on success, -1 on failure */
	if (!recno || recno > tablehandle->last_oid) {
		set_error("no record with such oid");
		return(-1);
	}
	/* iterate over list of all entries until right oid is encountered.. this is of course
	 * very inefficient, but can't really do it much better without implementing indexes
	 */
	while (1) {
		fread(&length, sizeof(int), 1, tablehandle->stream);
		fseek(tablehandle->stream, OID_OFFSET, SEEK_CUR);
		fread(&cur_oid, sizeof(long int), 1, tablehandle->stream);	
		if (recno == cur_oid) {
				
		}	
	}	
}

unsigned long int read_next(void **recbuf, struct table_handler *tablehandle) {
	unsigned int length;

#ifdef DEBUG
	printf("in read_next()\n");
#endif
	if (tablehandle->read_pos == 0) {
		/* nothing read yet.. first entry.. position indicator to the first entry */
		fseek(tablehandle->stream, tablehandle->start, SEEK_SET);
		fread(&length, sizeof(int), 1, tablehandle->stream);
		tablehandle->read_pos = tablehandle->start;
#ifdef DEBUG
		printf("FIRST read, length %d, repositioning indicators %ld\n", length, tablehandle->start);
#endif
	} else {
		fseek(tablehandle->stream, tablehandle->read_pos, SEEK_SET);
		fread(&length, sizeof(int), 1, tablehandle->stream);
#ifdef DEBUG
		printf("length of current entry %d position: %ld\n", length, tablehandle->read_pos);
#endif
		tablehandle->read_pos += (length + HEADER_SIZE );
		if (tablehandle->read_pos >= tablehandle->end) {
#ifdef DEBUG
			printf("reached end of DB. (%ld)\n", tablehandle->end);
#endif
			tablehandle->read_pos = 0;
			set_error("no more entries");
			return(0);	
		}
		fseek(tablehandle->stream, tablehandle->read_pos, SEEK_SET);
		fread(&length, sizeof(int), 1, tablehandle->stream);
#ifdef DEBUG
		printf("length of the entry to be read is %u\n", length);
#endif
	}
#ifdef DEBUG
	printf("struct size: %d\n", tablehandle->struct_size);
#endif
	*recbuf = (void*)malloc(tablehandle->struct_size);
#ifdef DEBUG
	printf("pointer to buffer is %ld\n", *recbuf);
#endif	
	read_in_rec(*recbuf, tablehandle);
}

int read_in_rec(void *recbuf, struct table_handler *tablehandle) {
	/* reads record from the current position of the tablehandle->cur_pos into recbuf */
	struct field *p;
	char *tmp;
	char c_len;
	int length;
	
	fseek(tablehandle->stream, tablehandle->read_pos + HEADER_SIZE , SEEK_SET);

	p = tablehandle->fieldlist;
	while (p) {
		switch (p->type) {
			case 'v':
				/* read in length of variable string */
				fread(&c_len, sizeof(char), 1, tablehandle->stream);
				c_len++; /* for \0 */
				/* allocate memory and copy the string there */
				tmp = malloc(c_len*sizeof(char));
				tmp[c_len-1] = 0;
				fread(tmp, sizeof(char), c_len-1, tablehandle->stream);	
				
#ifdef DEBUG
				printf("read varchar field with value %s\n", tmp);
#endif
				/* place pointer to the data structure for user */
				memcpy(recbuf, &tmp, sizeof(char*));
				/* advance position of the record buffer accordingly */
				recbuf += sizeof(char*);
				break;	
			case 'i':
				fread(recbuf, sizeof(int), 1, tablehandle->stream);
				recbuf += sizeof(int);
				break;
			case 'd':
				fread(recbuf, sizeof(double), 1, tablehandle->stream);
				recbuf += sizeof(double);
				break;
			case 'f':
				fread(recbuf, sizeof(float), 1, tablehandle->stream);
				recbuf += sizeof(float);
				break;
			case 't':
				fread(recbuf, sizeof(long), 1, tablehandle->stream);
				recbuf += sizeof(long);
				break;
		}
		p = p->next;
	}
	return(0);
}	

int seek_record(unsigned long int recno, struct table_handler *tablehandle) {
	/* moves file position indicator to the right record */

	

}

int show_error(void) {
	if (_errstr)
		printf("%s\n", _errstr);
}

int set_error(char *errstr) {
	_errstr = strdup(errstr);
}
